#!/usr/bin/perl -w
# 
# ftpsite.pl - 更新されたファイルを FTP で送る。
# Copyright 2003-2006 M.Shibata.
# http://www.emptypage.jp/gadget/ftpsite.html

use utf8;
use strict;

use Getopt::Long;
use Net::FTP;
use URI;
use URI::Escape;

my $OS_SEP = '/';
my $PERL_SEP = '/';
if ($^O eq 'MSWin32') { $OS_SEP = '\\'; }


my $CONFIG_FILENAME     = '.ftpsite';
my $TIMESTAMP_FILENAME  = '.ftpsite.log';
my $DEBUG   = 0;
my $VERBOSE = 0;
my $USAGE = <<END_OF_STRING;
ftpsite.pl [options] [directory]

options:
  --all         : transfer all specified files.
  --noftp       : show files that will be transfered.
  --verbose     : wordy messages to stderr.
  --status      : show modification status of files.
  --timestamp   : show modified times of files.
  --help        : show this help.

<mshibata\@emptypage.jp>
END_OF_STRING


sub perl_path {
    # OS 固有のパス区切り文字を Perl 内部での文字 ('/') に変換する。
    my ($path) = @_;
    $path =~ s/\Q$OS_SEP\E/$PERL_SEP/g;
    $path;
}


sub os_path {
    # Perl 内部でのパス区切り文字 ('/') を OS 固有の文字に変換する。
    my ($path) = @_;
    $path =~ s/\Q$PERL_SEP\E/$OS_SEP/g;
    return $path;
}


sub join_path {
    # パスを連結する。
    my @paths = ();
    foreach (grep {$_} @_) {
        @paths = () if /^\//;
        push @paths, ($_);
    }
    my $path = join '/', @paths;
    $path =~ s/\/+/\//;
    return $path;
}


sub split_path {
    # パスをディレクトリ名部とファイル名部に分割する。
    my ($path) = @_;
    my $sep = rindex $path, $PERL_SEP;
    my ($head, $tail);
    if ($sep < 0) {
        $head = '';
        $tail = $path;
    }
    if ($sep == 0) {
        $head = $PERL_SEP;
        $tail = substr $path, 1;
    }
    if (0 < $sep) {
        $head = substr $path, 0, $sep;
        $tail = substr $path, $sep + 1;
    }
    return ($head, $tail);
}


sub input_stdin {
    # $message を表示して標準入力から一行読む。取得文字列の末尾改行を削除した
    # ものを返す。
    my ($message) = @_;
    my $input;
    print $message;
    chomp ($input = <STDIN>);
    return $input;
}


sub print_verbose {
    my ($msg) = @_;
    print STDERR $msg if $VERBOSE;
}


sub print_debug {
    my ($msg) = @_;
    print STDERR $msg if $DEBUG;
}


sub load_config {
    # 設定ファイル $filename を読み込んで、ホスト名 $$r_remotepath と転送ファ
    # イルの指定リスト @$r_directives を埋める。
    my ($filename, $r_remotepath, $r_directives) = @_;
    open IN, $filename or die "can't open $filename: $!\n";
    while (<IN>) {
        s/^\s+|\s+$//;
        next if /^$|^#/;
        if (s/^\[(.+)\]$/$1/) {
            $$r_remotepath = $_;
            last;
        } else {
            die "remotehost format error ($filename): $_\n";
        }
    }
    while (<IN>) {
        s/^\s+|\s+$//;
        next if /^$|^#/;
        push @$r_directives, $_;
    }
    close IN;
}


sub load_timestamp {
    # 更新時刻ログファイル $filename を読み込んで、ファイルのパスと更新時刻画
    # 格納されたハッシュを返す。
    my ($filename) = @_;
    my %timestamp;
    if (open IN, $filename) {
        my ($k, $v);
        while (<IN>) {
            chomp;
            ($k, $v) = split /\t/;
            $timestamp{$k} = $v;
        }
        close IN;
    }
    return %timestamp;
}


sub glob_files {
    # パスマスク指定からファイルのリストを返す。$deepが真の場合はサブフォルダ
    # も含める。
    my ($dir, $mask, $deep) = (@_);
    $mask = '*' if not defined $mask;
    my @files;
    my @dirs = ($dir);
    while (@dirs) {
        $dir = shift @dirs;
        push @files, grep {not -d} glob join_path($dir, $mask);
        push(@dirs, grep {-d} glob join_path($dir, '*')) if $deep;
    }
    return @files;
}


sub get_timestamps {
    my ($localpath, @directives) = @_;
    # $localpath を基準として、@directives 指定に当てはまるファイルのパスをリ
    # ストで返す。
    my ($directive, $remove, $dir, $mask, $deep, @files, $filename);
    my %count;
    foreach $directive (@directives) {
        # '!' で始まる指定は除外するファイル
        $remove = ($directive =~ s/^\!\s*//);
        # 指定を通常部分と拡張部分 (<...>) とに分ける。
        ($dir, $mask) = split /(<[^>]+>)/, $directive;
        if ($mask) {
            $deep = 1;
            $mask =~ s/^<|>$//g;
        } else {
            ($dir, $mask) = split_path($dir);
        }
        @files = glob_files(join_path($localpath, $dir), $mask, $deep);
        if ($remove) {
            delete @count{@files};
            do {print STDERR "- $_\n" foreach @files} if $DEBUG;
        } else {
            @count{@files} = @files;
            do {print STDERR "+ $_\n" foreach @files} if $DEBUG;
        }
    }
    my (%timestamp, $key);
    foreach (keys %count) {
        ($key = $_) =~ s/^\Q$localpath\E\///;
        $timestamp{$key} = (stat)[9];
    }
    return %timestamp;
}


sub compare_status {
    # ある名前についてその状態を記録したハッシュふたつを比較する。
    # $data1, $data2 は状態1, 2 を保持するハッシュへの参照。
    # 比較結果として名前について以下の文字を格納したハッシュを返す。
    # ' '   変更なし。
    # 'A'   追加された。
    # 'D'   削除された。
    # 'M'   変更された。
    my ($data1, $data2) = @_;
    my %result;
    my ($name, $status1, $status2);
    foreach $name (keys %$data1) {
        $status1 = $$data1{$name};
        $status2 = $$data2{$name};
        if (defined $status2) {
            $result{$name} = ($status1 == $status2) ? ' ' : 'M';
        } else {
            $result{$name} = 'D';
        }
    }
    foreach $name (keys %$data2) {
        $result{$name} = 'A' if not defined $result{$name};
    }
    return %result;
}


sub ftp_files {
    # ファイルを FTP で転送する。
    # $localroot: 基準ディレクトリ。
    # $hostname, $remoteroot, $username, $password, $pasv: 転送用情報。
    # @files: 基準ディレクトリからの相対指定でのファイル一覧。
    my ($localroot, $hostname, $remoteroot,
        $username, $password, $pasv, @files) = @_;
    print_verbose("Connecting to: $hostname\n");
    my $ftp = Net::FTP->new($hostname, Passive => $pasv)
        or die "Can't connect to $hostname: $!\n";
    print_verbose("Logging in: $username\n");
    $ftp->login($username, $password) or die "failed: $!\n";
    foreach (grep {$_} split('/', $remoteroot)) {
        $ftp->cwd(uri_unescape($_)) or die "Can't go to $_: $!\n";
    }
    $ftp->binary();
    my @succeeded;
    foreach (@files) {
        if ($ftp->put(join_path($localroot, $_), $_)) {
            print_verbose("  $_\n");
            push @succeeded, $_;
        } else {
            print_verbose("!  $_\n");
        }
    }
    $ftp->quit();
    return @succeeded;
}


sub print_status {
    # {キー => 状態} の辞書をタブ区切りで書き出す。
    # $all が指定されているときは、状態が空白 (' ') でも表示する。
    my ($filestate, $all) = @_;
    my $status;
    foreach (keys %$filestate) {
        $status = $$filestate{$_};
        print"$status\t$_\n" if $all or $status ne ' ';
    }
}


sub print_files {
    my ($localpath, $filestate, $all) = @_;
    my ($path, $status);
    foreach (keys %$filestate) {
        $path = os_path(join_path($localpath, $_));
        $status = $$filestate{$_};
        print "$path\n" if $all or $status eq 'A' or $status eq 'M';
    }
}


sub print_timestamp {
    my ($timestamp, $fh) = @_;
    foreach (keys %$timestamp) {
        print $fh "$_\t$$timestamp{$_}\n";
    }
}


sub main {
    my ($opt_all, $opt_noftp, $opt_status, $opt_help, $opt_timestamp);
    GetOptions(
        'all'       => \$opt_all,
        'noftp'     => \$opt_noftp,
        'status'    => \$opt_status,
        'timestamp' => \$opt_timestamp,
        'help|?'    => \$opt_help,
        'debug'     => \$DEBUG,
        'verbose'   => \$VERBOSE,
    ) or die $USAGE;
    if ($opt_help) { print $USAGE; return }
    
    my $localpath = perl_path(shift(@ARGV) or "");
    my $config_filename     = join_path($localpath, $CONFIG_FILENAME);
    my $timestamp_filename  = join_path($localpath, $TIMESTAMP_FILENAME);
    
    my ($remotepath, @directives);
    print_verbose("Loading: $config_filename\n");
    load_config($config_filename, \$remotepath, \@directives);
    
    my %timestamp_old = ();
    if (-e $timestamp_filename) {
        print_verbose("Loading: $timestamp_filename\n");
        %timestamp_old = load_timestamp($timestamp_filename);
    }
    print_verbose("Crawling: $localpath\n");
    my %timestamp_new = get_timestamps($localpath, @directives);
    my %status = compare_status(\%timestamp_old, \%timestamp_new);
    
    if ($opt_noftp)     { print_files($localpath, \%status, $opt_all); return }
    if ($opt_status)    { print_status(\%status, $opt_all); return }
    if ($opt_timestamp) { print_timestamp(\%timestamp_new, *STDOUT); return }
    
    my @modified_files = grep {0 <= index('AM', $status{$_})} keys %status;
    if (not @modified_files) { print_verbose("No modified file.\n"); return }
    
    my $uri = URI->new($remotepath);
    my @succeeded = ftp_files($localpath, $uri->host, $uri->path, $uri->user,
        ($uri->password or input_stdin("password: ")), 1, @modified_files);
    if (@succeeded) {
        $status{$_} = ' ' foreach @succeeded;
        delete @timestamp_new{grep {$status{$_} ne ' '} keys %timestamp_new};
        
        print_verbose("Saving: $timestamp_filename\n");
        open FH, '>', $timestamp_filename
            or die "can't open $timestamp_filename: $!\n";
        print_timestamp(\%timestamp_new, *FH);
        close FH;
    }
}

main();
